"use strict";
/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        if (typeof b !== "function" && b !== null)
            throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
exports.shadowBlockConversionChangeListener = exports.BlockShadowChange = void 0;
/**
 * @fileoverview Plugin for converting shadow blocks to real ones on edit.
 */
var Blockly = require("blockly/core");
/**
 * A new blockly event class specifically for recording changes to the shadow
 * state of a block. This implementation is similar to and could be merged with
 * the implementation of Blockly.Events.BlockChange in Blockly core code.
 */
var BlockShadowChange = /** @class */ (function (_super) {
    __extends(BlockShadowChange, _super);
    /**
     * The constructor for a new BlockShadowChange event.
     * @param block The changed block. Undefined for a blank event.
     * @param oldValue Previous value of shadow state.
     * @param newValue New value of shadow state.
     */
    function BlockShadowChange(block, oldValue, newValue) {
        var _this = _super.call(this, block) || this;
        _this.type = BlockShadowChange.EVENT_TYPE;
        if (!block) {
            return _this;
        }
        _this.oldValue = typeof oldValue === 'undefined' ? '' : oldValue;
        _this.newValue = typeof newValue === 'undefined' ? '' : newValue;
        return _this;
    }
    /**
     * Encode the event as JSON.
     * @returns JSON representation.
     * @override
     */
    BlockShadowChange.prototype.toJson = function () {
        var json = _super.prototype.toJson.call(this);
        json['oldValue'] = this.oldValue;
        json['newValue'] = this.newValue;
        return json;
    };
    /**
     * Decode the JSON event.
     * @param json JSON representation.
     * @override
     */
    BlockShadowChange.fromJson = function (json, workspace, event) {
        var newEvent = _super.fromJson.call(this, json, workspace, event);
        newEvent.oldValue = json['oldValue'];
        newEvent.newValue = json['newValue'];
        return event;
    };
    /**
     * Does this event record any change of state?
     * @returns False if something changed.
     * @override
     */
    BlockShadowChange.prototype.isNull = function () {
        return this.oldValue === this.newValue;
    };
    /**
     * Run a change event.
     * @param forward True if run forward, false if run backward (undo).
     * @override
     */
    BlockShadowChange.prototype.run = function (forward) {
        var workspace = this.getEventWorkspace_();
        if (!this.blockId) {
            throw new Error('The block ID is undefined. Either pass a block to ' +
                'the constructor, or call fromJson');
        }
        var block = workspace.getBlockById(this.blockId);
        if (!block) {
            throw new Error('The associated block is undefined. Either pass a ' +
                'block to the constructor, or call fromJson');
        }
        var value = forward ? this.newValue : this.oldValue;
        block.setShadow(!!value);
    };
    /**
     * The name of the event type for broadcast and listening purposes.
     */
    /* eslint-disable @typescript-eslint/naming-convention */
    BlockShadowChange.EVENT_TYPE = 'block_shadow_change';
    return BlockShadowChange;
}(Blockly.Events.BlockBase));
exports.BlockShadowChange = BlockShadowChange;
Blockly.registry.register(Blockly.registry.Type.EVENT, BlockShadowChange.EVENT_TYPE, BlockShadowChange);
/**
 * Add this function to your workspace as a change listener to automatically
 * convert shadow blocks to real blocks whenever the user edits a field on the
 * block, like this:
 *
 * workspace.addChangeListener(shadowBlockConversionChangeListener);
 *
 * Ideally the Blockly.Field.prototype.setValue method should handle this logic,
 * but for the purposes of this plugin it can be a workspace change listener.
 *
 * @param event An event broadcast by the workspace.
 */
function shadowBlockConversionChangeListener(event) {
    var _a, _b;
    // Auto-converting shadow blocks to real blocks should happen in response to
    // new user action events (which get recorded as undo events) but not when
    // undoing or redoing events (which do not get recorded again).
    if (!event.recordUndo) {
        return;
    }
    // Auto-converting shadow blocks to real blocks should happen in response to
    // editing a field value, which is recorded as Blockly.Events.BLOCK_CHANGE.
    if (event.type != Blockly.Events.BLOCK_CHANGE) {
        return;
    }
    var blockEvent = event;
    if (!blockEvent.workspaceId || !blockEvent.blockId) {
        return;
    }
    var workspace = Blockly.Workspace.getById(blockEvent.workspaceId);
    if (!workspace) {
        return;
    }
    var block = workspace.getBlockById(blockEvent.blockId);
    if (!block) {
        return;
    }
    // Blocks that are already real blocks can be ignored.
    if (!block.isShadow()) {
        return;
    }
    // Remember the current event group so that it can be resumed below.
    var currentGroup = Blockly.Events.getGroup();
    if (blockEvent.group) {
        // Temporarily use the same group as the initiating event so that
        // the shadow events get grouped with it for undo purposes.
        Blockly.Events.setGroup(blockEvent.group);
    }
    else {
        // The initiating event wasn't part of any named group, so the shadow events
        // can't be grouped with it, but at least they can be grouped with each
        // other.
        Blockly.Events.setGroup(true);
    }
    // If the changed shadow block is a child of another shadow block, then both
    // blocks should be converted to real blocks. To find all the shadow block
    // ancestors that need to be converted to real blocks, seed the list of blocks
    // starting with the changed block, and append all shadow block ancestors.
    var shadowBlocks = [block];
    for (var i = 0; i < shadowBlocks.length; i++) {
        var shadowBlock = shadowBlocks[i];
        // If connected blocks need to be converted too, add them to the list.
        var outputBlock = (_a = shadowBlock.outputConnection) === null || _a === void 0 ? void 0 : _a.targetBlock();
        var previousBlock = (_b = shadowBlock.previousConnection) === null || _b === void 0 ? void 0 : _b.targetBlock();
        if (outputBlock === null || outputBlock === void 0 ? void 0 : outputBlock.isShadow()) {
            shadowBlocks.push(outputBlock);
        }
        if (previousBlock === null || previousBlock === void 0 ? void 0 : previousBlock.isShadow()) {
            shadowBlocks.push(previousBlock);
        }
    }
    // The list of shadow blocks starts with the deepest child and ends with the
    // highest parent, but the parent of a real block should never be a shadow
    // block, so the parents need to be converted to real blocks first. Start
    // at the end of the list and iterate backward to convert the blocks.
    for (var i = shadowBlocks.length - 1; i >= 0; i--) {
        var shadowBlock = shadowBlocks[i];
        // Convert the shadow block to a real block and fire an event recording the
        // change so that it can be undone. Ideally the
        // Blockly.Block.prototype.setShadow method should fire this event directly,
        // but for this plugin it needs to be explicitly fired here.
        shadowBlock.setShadow(false);
        Blockly.Events.fire(new BlockShadowChange(shadowBlock, true, false));
    }
    // Revert to the current event group, if any.
    Blockly.Events.setGroup(currentGroup);
}
exports.shadowBlockConversionChangeListener = shadowBlockConversionChangeListener;
